import sys
from os.path import dirname as d
from os.path import abspath, join, pathsep

root_dir = d(d(abspath(__file__)))
env = join(root_dir, ".env")
try:
    with open(env) as f:

        def ispypath(s: str) -> bool:
            return s.startswith("PYTHONPATH")

        pythonpaths = [x[11:].rstrip() for x in f.readlines() if ispypath(x)]
        if pythonpaths:
            pythonpath = pythonpaths[0]
            pythonpath = pythonpath.replace("${env:PROJ_DIR}", root_dir)
            pythonpaths = pythonpath.split(pathsep)
            for p in pythonpaths:
                sys.path.append(p)
except FileNotFoundError:
    # The env file is not to be found, this is a production build.
    pass
sys.path.append(join(root_dir, "src"))



import logging
import time


from typing import Tuple, List

import grpc

import blindecdh
import pskca

from cakes.proto import cakes_pb2_grpc as pb2_grpc, cakes_pb2 as pb2
from cakes.types import (
    ECDHVerificationCallback,
    CertificateIssuedCallback,
    RejectedByPeer,
    RejectedBySelf,
    Ignored,
    EPERM,
    EWAIT,
)
from cakes.util import (
    unconditional_accept_cert,
    unconditional_accept_ecdh,
)

from cryptography.x509 import (
    Certificate,
    CertificateSigningRequest,
)
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.serialization import (
    load_pem_public_key,
)

_LOGGER = logging.getLogger(__name__)


class CAKESClient(object):
    def __init__(
        self,
        channel: grpc.Channel,
        csr: CertificateSigningRequest,
        verification_callback: ECDHVerificationCallback,
        certificate_issued_callback: CertificateIssuedCallback,
    ) -> None:
        """Initializes the CAKES client.

        Parameters:
            channel: a grpc.Channel to talk to the server through.
            csr: a certificate signing request to be signed by the
            CAKES server.
            verification_callback: a callback which the ECDH exchange process
            will call to give your code an opportunity to accept or reject
            the result of an ECDH exchange.  The process does not move forward
            until your callback returns True.  If it returns False, the
            exception RejectedBySelf is raised in run().
            certificate_issued_callback: a callback issued prior to returning
            from the process (see function run()) that gives your code the
            opportunity to accept or reject the certificate.  If it returns
            False, the exception RejectedBySelf is raised in run().  This
            callback receives an empty peer name, the issued certificate, and
            a list (chain) of certificates that form the root of trust used
            to validate the identity of the authenticated server.
        """
        self.channel = channel
        self.csr = csr
        self.verification_callback = verification_callback
        self.cert_issued_callback = certificate_issued_callback

    def run(
        self,
        retry_interval: float = 3.0,
        deadline: float = 60.0,
    ) -> Tuple[Certificate, List[Certificate]]:
        """Runs the CAKES protocol against the server connected to the channel.

        If the server provisionally accepts the ECDH exchange, then the
        verification_callback will be called (with an empty peer name and the
        completed ECDH exchange).  The callback can then decide whether the
        exchange is verified, and return True -- or reject the exchange,
        in which case it must return False (then this raises RejectedBySelf).

        At the end of a successful process, it returns the certificate issued
        to your code, as well as the certificate chain for the root of trust
        you can use to verify the authenticated server.

        If the server rejected the ECDH exchange or otherwise failed to perform
        the exchange or issue the cert, an exception RejectedByPeer is raised.

        If the certificate issuance callback return False, the exception
        RejectedBySelf is raised.

        If the communication was tampered between the parties, an exception
        pskca.CannotDecrypt will be raised.

        Parameters:
            retry_interval: (default 3) how many seconds to wait between
            retries when the server says that verification is still pending.
        """
        s = blindecdh.ECDHProtocol()
        stub = pb2_grpc.CAKESStub(self.channel)  # type: ignore
        try:
            stub.ClientPubkey(
                pb2.ECDHKey(
                    pubkey=s.public_key.public_bytes(
                        serialization.Encoding.PEM,
                        serialization.PublicFormat.SubjectPublicKeyInfo,
                    )
                )
            )
            reply = stub.ServerPubkey(pb2.Ack()).pubkey
            remote_pubkey_pem = load_pem_public_key(reply)
        except grpc.RpcError as e:
            if e.code() == EPERM:
                raise RejectedByPeer(e.details())
            raise

        complete = s.run(remote_pubkey_pem)

        result = self.verification_callback("", complete)
        if not result:
            raise RejectedBySelf()

        psk = complete.derived_key

        requestor = pskca.Requestor(psk)

        request = pb2.IssueCertificateRequest(
            EncryptedCSR=requestor.encrypt_csr(self.csr).to_bytes()
        )

        start = time.time()
        retry = False
        while True:
            _LOGGER.debug("Attempting to request certificate.")
            try:
                reply = stub.IssueCertificate(request)
                retry = False
            except grpc.RpcError as e:
                if e.code() == EPERM:
                    raise RejectedByPeer(e.details())
                elif e.code() == EWAIT:
                    _LOGGER.debug("The server says we are not authorized yet.")
                    elapsed = time.time() - start
                    if elapsed >= deadline:
                        raise Ignored(
                            "The server did not authorize us in %.2f seconds"
                            % (elapsed,)
                        )
                    else:
                        retry = True
                else:
                    raise
            if retry:
                time.sleep(retry_interval)
            else:
                break

        cert, chain = requestor.decrypt_reply(
            pskca.EncryptedClientCertificate(reply.EncryptedClientCert),
            pskca.EncryptedCertificateChain(reply.EncryptedCertChain),
        )

        result = self.cert_issued_callback("", cert, chain)
        if not result:
            raise RejectedBySelf()

        return cert, chain


__all__ = ["CAKESClient"]


def __client() -> None:
    csr, unused_key = pskca.create_certificate_signing_request()
    with grpc.insecure_channel("127.0.40.50:50052") as channel:
        client = CAKESClient(
            channel,
            csr,
            unconditional_accept_ecdh,
            unconditional_accept_cert,
        )
        print("connecting client")
        one, two = client.run()
        print(one)
        print(two)
        print("done")


if __name__ == "__main__":
    __client()



"""
Certificate authority kit, ECDH-powered, SAS-authenticated.

This package is a kit to build mutually-authenticated secure services.
You will find this package useful if:

* You have a server program, and a client program.
* You want to secure communications between these programs using SSL
  or gRPC secure channels.
* You do not have a certificate authority or other PKI system on hand.
* All you need is a way to authenticate the client on the server side,
  and the server on the client side.
* You want to authenticate the two parts by making a visual comparison
  on both sides (much like Bluetooth PIN pairing).

The way you use this goes more or less as explained here:

Step 1: make your server run a CAKESServicer on an insecure gRPC channel.

```
# Create a dummy CA cert with the power to issue certificates.
# You can use an actual CA cert or a CA cert delegated by a
# root cert.
ca_cert, ca_key = pskca.create_certificate_and_key(
    ca=True,
    cn="dummy.not.for.production"
)
# Create the CA object that will issue certs.
# The third parameter is going to be the chain of certificates that
# the client will be given to trust, so it does not have to be the
# CA cert itself.
ca = pskca.CA(ca_cert, ca_key, [ca_cert])
# Create your gRPC server.
server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
# Create the CAKESServicer and add some callbacks to authenticate
pb2_grpc.add_CAKESServicer_to_server(
    cakes.CAKESServicer(
        ca,
        accept_ecdh,
        cakes.unconditional_accept_cert,
    ),
    server,
)
# Add insecure port to your server.
server.add_insecure_port("0.0.0.0:50052")
# Start and wait.
server.start()
server.wait_for_termination()
```

Step 2: make your client adopt a CAKESClient to obtain its certificate.

```
# Generate dummy certificate signing request.
csr, client_key = pskca.create_certificate_signing_request()
# Create gRPC insecure channel.
with grpc.insecure_channel("localhost:50052") as channel:
    client = cakes.CAKESClient(
        channel,
        csr,
        accept_ecdh,
        cakes.unconditional_accept_cert,
    )
    client_cert, chain = client.run()
```

Step 3: provide an authentication function.

You'll note that in both steps above there's an `accept_ecdh` function in use.
You supply this function; both the client and the server will call it with
the name of the peer, and an object of class blindecdh.CompletedECDH.  This
object has an attribute `derived_key` that is a 32-byte array.  You can then
make your `accept_ecdh` function print or display a simplified version of the
key, and ask if the exchange is accepted:

```
def accept_ecdh(peer, completed_ecdh):
    from shortauthstrings import emoji
    s = emoji(completed_ecdh.derived_key)
    print("Peer %s says he has key %s.  Accept?  [Y/N then ENTER]" % (peer, s))
    response = sys.stdin.readline()
    if response.lower().startswith("y"):
        return True
    return False
```

Step 4: use the CA / signing certificate on the server

Once the client has received its certificate, then the server and the client
can talk in the customary way.  The server needs to use a certificate either
signed by the CA you used in step 1, or the same certificate.  Effectively,
the server has to use a certificate that is in the chain of trust given to
the client in step #2.

```
# server side
...
credentials = grpc.ssl_server_credentials(
    (
        (
            ca_cert.public_bytes(serialization.Encoding.PEM),
            ca_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            ),
        ),
    ),
    PEM.from_rsa_certificate(root_cert).as_bytes(),
    require_client_auth=True,
)
server.add_secure_port("0.0.0.0:50051", credentials)
...
```

Step 5: use the obtained certificate on the client.

```
# client side
pem_chain = "\n".join([
    c.public_bytes(serialization.Encoding.PEM)
    for c in chain
])
pem_client_key = client_key.private_bytes(
    encoding=serialization.Encoding.PEM,
    format=serialization.PrivateFormat.TraditionalOpenSSL,
    encryption_algorithm=serialization.NoEncryption(),
)
pem_client_cert = client_cert.public_bytes(
    serialization.Encoding.PEM
)
credentials = grpc.ssl_channel_credentials(
    pem_chain,
    pem_client_key,
    pem_client_cert,
)
# We must override the target name so the server chooses
# the right certificate to present to us, irrespective
# of the actual DNS or IP address used.  Since we forced the CN
# of the CA certificate, and we're reusing the CA certificate as
# the server certificate, this will work.
server_cert_hostname = "dummy.not.for.production"
server_hostname = "127.0.0.1"
channel = grpc.secure_channel(
    "%s:50051" % server_hostname,
    credentials,
    options=[("grpc.ssl_target_name_override", server_cert_hostname)],
)
stub = myservice_pb2_grpc.MyServiceServiceStub(channel=channel)
# This call is now encrypted using SSL and mutually authenticated.
stub.my_function()
...
```

How does this work?  It's fairly simple.

1. The client talks over the insecure port to the server, initiating
   an ECDH exchange.
2. The server cooperates with the client in the ECDH exchange.
3. Both sides ask the user for authentication, which uses a visually
   easy-to-use representation (we recommend using package `shortauthstrings`)
   to display to the user what the resulting ECDH key was.
4. Once both sides are verified, we know there's no man-in-the-middle
   attack, and the second part happens.  The client encrypts its CSR
   using the agreed key and sends it to the server.
5. The server decrypts the CSR with the agreed key, and signs it, returning
   to the client an encrypted certificate as well as an encrypted chain
   of trust.
6. The client receives these values, then decrypts them.
7. Server and client may now begin using these newly-agreed-upon certificate
   materials to do actual mutually-authenticated SSL.
   (Technically speaking, if the CA cert is used for the CA object, and the
   CA cert also issues the server certificate, the secure service may start
   -- using the server certificate -- before any client certificate is
   issued.  So long as the chain of trust passed to the CA object is correct,
   then any client which later gets certificates issued should just work by
   using those certificates issued to them.)
"""

# FIXME: we also need a schematic of how the thing works and how the
# certs relate to each other.

from cakes.client import CAKESClient  # noqa: F401
from cakes.server import CAKESServicer  # noqa: F401
from cakes.util import (  # noqa: F401
    unconditional_accept_cert,
    unconditional_accept_ecdh,
    accept_ecdh_via_console,
)

from cakes.types import (  # noqa: F401
    CannotDecrypt,
    RejectedBySelf,
    Rejected,
    RejectedByPeer,
    Ignored,
    ECDHVerificationCallback,
    CertificateIssuedCallback,
)

__version__ = "0.1.4"

__all__ = [
    "unconditional_accept_cert",
    "unconditional_accept_ecdh",
    "accept_ecdh_via_console",
    "CannotDecrypt",
    "Rejected",
    "RejectedBySelf",
    "RejectedByPeer",
    "Ignored",
    "ECDHVerificationCallback",
    "CertificateIssuedCallback",
    "CAKESClient",
    "CAKESServicer",
]



import grpc
from typing import Callable, List
import blindecdh
import pskca
from cryptography.x509 import Certificate


ECDHVerificationCallback = Callable[[str, blindecdh.CompletedECDH], bool]
CertificateIssuedCallback = Callable[
    [str, Certificate, List[Certificate]],
    bool,
]


class Rejected(Exception):
    """Base class for rejections."""


class RejectedByPeer(Rejected):
    """
    This exception is raised when the client code runs the ECDH exchange,
    and the server callback rejected the exchange.
    """

    pass


class RejectedBySelf(Rejected):
    """
    This exception is raised when the client code runs the ECDH exchange,
    and the client callback rejected the exchange.
    """

    pass


class Ignored(Rejected):
    """
    This exception is raised when the server never authorized the ECDH
    exchange in the deadline given to the CAKESCLient.run() method.
    """


Pending = pskca.Pending

CannotDecrypt = pskca.CannotDecrypt


EPERM = grpc.StatusCode.PERMISSION_DENIED
EWAIT = grpc.StatusCode.UNAUTHENTICATED
EINVAL = grpc.StatusCode.INVALID_ARGUMENT



import sys
from typing import List
import blindecdh
from cryptography.x509 import Certificate


def unconditional_accept_ecdh(
    peer: str,
    unused: blindecdh.CompletedECDH,
) -> bool:
    """
    Stub callback function to auto-accept an ECDH exchange.

    Do not use in production.  Blindly "verifying" an ECDH exchange without
    comparing the two CompleteECDH exchanges on both sides makes your code
    vulnerable to active man-in-the-middle attacks.
    """
    return True


def unconditional_accept_cert(
    peer: str,
    unused: Certificate,
    unused_list: List[Certificate],
) -> bool:
    """
    Stub callback function to auto-accept an issued certificate.

    You may use this in production.
    """
    return True


def accept_ecdh_via_console(
    peer: str,
    complete: blindecdh.CompletedECDH,
) -> bool:
    """
    Sample function to accept ECDHs via terminal standard input/output.

    Your CAKES servicer will ask you on the terminal if you use this function
    as verification callback for ECDHs.

    This is useful for prototyping services on a console, but should not be
    used in production.
    """
    print("Peer says it's %s" % peer)
    print("Key appears to be %r" % complete.derived_key)
    print("Accept?  [Y/N then ENTER]")
    line = sys.stdin.readline()
    result = line.lower().startswith("y")
    return result


__all__ = [
    "unconditional_accept_ecdh",
    "unconditional_accept_cert",
    "accept_ecdh_via_console",
]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cakes.proto import cakes_pb2 as cakes__pb2


class CAKESStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ClientPubkey = channel.unary_unary(
                '/CAKES.CAKES/ClientPubkey',
                request_serializer=cakes__pb2.ECDHKey.SerializeToString,
                response_deserializer=cakes__pb2.Ack.FromString,
                )
        self.ServerPubkey = channel.unary_unary(
                '/CAKES.CAKES/ServerPubkey',
                request_serializer=cakes__pb2.Ack.SerializeToString,
                response_deserializer=cakes__pb2.ECDHKey.FromString,
                )
        self.IssueCertificate = channel.unary_unary(
                '/CAKES.CAKES/IssueCertificate',
                request_serializer=cakes__pb2.IssueCertificateRequest.SerializeToString,
                response_deserializer=cakes__pb2.IssueCertificateReply.FromString,
                )


class CAKESServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ClientPubkey(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ServerPubkey(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IssueCertificate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_CAKESServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ClientPubkey': grpc.unary_unary_rpc_method_handler(
                    servicer.ClientPubkey,
                    request_deserializer=cakes__pb2.ECDHKey.FromString,
                    response_serializer=cakes__pb2.Ack.SerializeToString,
            ),
            'ServerPubkey': grpc.unary_unary_rpc_method_handler(
                    servicer.ServerPubkey,
                    request_deserializer=cakes__pb2.Ack.FromString,
                    response_serializer=cakes__pb2.ECDHKey.SerializeToString,
            ),
            'IssueCertificate': grpc.unary_unary_rpc_method_handler(
                    servicer.IssueCertificate,
                    request_deserializer=cakes__pb2.IssueCertificateRequest.FromString,
                    response_serializer=cakes__pb2.IssueCertificateReply.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'CAKES.CAKES', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class CAKES(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ClientPubkey(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/CAKES.CAKES/ClientPubkey',
            cakes__pb2.ECDHKey.SerializeToString,
            cakes__pb2.Ack.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ServerPubkey(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/CAKES.CAKES/ServerPubkey',
            cakes__pb2.Ack.SerializeToString,
            cakes__pb2.ECDHKey.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def IssueCertificate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/CAKES.CAKES/IssueCertificate',
            cakes__pb2.IssueCertificateRequest.SerializeToString,
            cakes__pb2.IssueCertificateReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import logging
import threading

from typing import Optional
import blindecdh
import pskca

import grpc
from cakes.proto import cakes_pb2_grpc as pb2_grpc, cakes_pb2 as pb2
from cakes.types import (
    ECDHVerificationCallback,
    CertificateIssuedCallback,
    EPERM,
    EWAIT,
    EINVAL,
)
from cakes.util import (
    unconditional_accept_cert,
    unconditional_accept_ecdh,
)


from pskca.util import TimedDict

from cryptography.hazmat.primitives import serialization

from cryptography.hazmat.primitives.serialization import (
    load_pem_public_key,
)
from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePublicKey,
)


_LOGGER = logging.getLogger(__name__)

# FIXME prevent abuse by forcing the client to do
# an expensive computation.


class CAKESServicer(pb2_grpc.CAKESServicer):
    def __init__(
        self,
        ca: pskca.CA,
        verification_callback: ECDHVerificationCallback,
        certificate_issued_callback: CertificateIssuedCallback,
    ):
        """
        Initializes the CAKES servicer.

        This gRPC servicer provides services to CAKES clients.

        All you must bring to the table is a callback function that, when
        called, shows an authorization request to the user, and if responded
        in the affirmative by him, returns True.

        Parameters:
            ca: a `pskca.CA` object.
            verification_callback: a callback to call with the peer
            identification and the completed ECDH to further process down the
            line.  After completed ECDH, the callback function will be called,
            so your code must use that callback function to authenticate that
            the client identity matches (e.g. via short authentication strings,
            as per the shortauthstrings package).
            certificate_issued_callback: a callback that will be called with
            the peer and the issued certificate every time a certificate is
            issued.  If this callback does not return True, the certificate is
            denied to the client.
        """
        p: TimedDict[str, Optional[EllipticCurvePublicKey]] = TimedDict(60, 4)
        self.peers = p
        self.ca = ca
        self.callback = verification_callback
        self.cert_issued_callback = certificate_issued_callback

    def ClientPubkey(
        self, request: pb2.ECDHKey, context: grpc.ServicerContext
    ) -> pb2.Ack:
        """
        Retrieves the public key of the ECDH exchange from the client.

        This is the first step in CAKES.
        """
        peer: str = context.peer()
        _LOGGER.debug("Peer ClientPubkey %s", peer)
        with self.peers:
            if peer in self.peers:
                context.abort(EPERM, "an ECDH is already ongoing")
        k = load_pem_public_key(request.pubkey)
        if not isinstance(k, EllipticCurvePublicKey):
            context.abort(EINVAL, "not a valid public key")
        self.peers[peer] = k
        return pb2.Ack()

    def ServerPubkey(
        self, request: pb2.Ack, context: grpc.ServicerContext
    ) -> pb2.ECDHKey:
        """
        Sends the public key of the ECDH exchange to the client.

        This is the second step in CAKES.

        At the end of this process, after the public key has been
        returned to the client, the callback function is called
        (in a separate thread, so keep thread safety in mind).

        If that function returns True, then the client may invoke
        the next step in CAKES -- the IssueCertificate RPC call.
        """
        peer: str = context.peer()
        _LOGGER.debug("Peer ServerPubkey %s", peer)
        with self.peers:
            try:
                peer_pubkey = self.peers[peer]
                self.peers[peer] = None
            except KeyError:
                context.abort(EPERM, "unknown ECDH")
                return
        if peer_pubkey is None:
            context.abort(EPERM, "an ECDH is already ongoing")
            return

        s = blindecdh.ECDHProtocol()
        if not isinstance(
            peer_pubkey,
            s.public_key.__class__,
        ):
            context.abort(
                EINVAL,
                "the curve of the ECDH exchange is wrong (%s != %s)"
                % (peer_pubkey.__class__, s.public_key.__class__),
            )
            return
        complete = s.run(peer_pubkey)

        def accept(peer: str, complete: blindecdh.CompletedECDH) -> None:
            _LOGGER.debug("Adding PSK of peer %s to pending", peer)
            self.ca.add_psk(peer, None)
            if self.callback(peer, complete):
                _LOGGER.debug("Adding PSK of peer %s to completed", peer)
                self.ca.add_psk(peer, complete.derived_key)
            else:
                _LOGGER.debug("Ignoring PSK of peer %s", peer)

        t = threading.Thread(
            target=accept,
            args=(
                peer,
                complete,
            ),
            daemon=True,
        )
        t.start()
        return pb2.ECDHKey(
            pubkey=s.public_key.public_bytes(
                serialization.Encoding.PEM,
                serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        )

    def IssueCertificate(
        self,
        request: pb2.IssueCertificateRequest,
        context: grpc.ServicerContext,
    ) -> pb2.IssueCertificateReply:
        """
        Issues a certificate to authorized clients.

        A client is authorized if:

        1. It has completed the ECDH phase (ClientPubkey + ServerPubkey)
        2. During ServerPubkey, the callback attached to this class returned
           True.
        3. At the time of calling this, the certificate_issued_callback
           is called, and the return value of the call is True.  Note that
           the call is blocking, so the client will not receive a response
           until and unless the callback is finished.

        The client must call this with an encrypted payload.  The payload
        must contain its certificate request, and must be encrypted with the
        derived_key of the CompletedECDH object on its side, which it obtains
        after successfully completing the ServerPubkey phase of the CAKES
        protocol.

        A reply to authenticated clients includes a message with two payloads:

        1. The issued certificate.  A newly-issued certificate, signed by
           the `pskca.CA` object, which the caller may use to do mutual TLS
           via gRPC or mTLS to any services protected by certificates signed
           by the CA (or the CA certificate itself).
        2. A chain of trust (derived from the `pskca.CA` object) to be sent
           to the client for purposes of authenticating against the
           authenticated service that the issued certificate is for.

        Replies to unsuccessful clients generally involve the gRPC status code
        PERMISSION_DENIED.

        If a client has completed the ECDH phase but the server has not yet
        finished authorizing the client's ECDH, the client will receive a reply
        with gRPC status code UNAUTHENTICATED.  The client should retry in a
        few seconds.
        """
        peer: str = context.peer()
        try:
            cert, chain, enc_cert, enc_chain = self.ca.issue_certificate(
                peer,
                pskca.EncryptedCertificateRequest(request.EncryptedCSR),
            )
        except pskca.UnknownRequestor:
            context.abort(EPERM, "no corresponding ECDH")
            return
        except pskca.Pending:
            context.abort(
                EWAIT,
                "the user has not yet approved the ECDH exchange",
            )
            return
        except pskca.CannotDecrypt:
            context.abort(
                EPERM,
                "cannot decrypt certificate signing request",
            )
            return

        cb_return = self.cert_issued_callback(peer, cert, chain)
        if not cb_return:
            context.abort(EPERM, "server denied certificate")
            return

        response = pb2.IssueCertificateReply(
            EncryptedClientCert=enc_cert.to_bytes(),
            EncryptedCertChain=enc_chain.to_bytes(),
        )
        _LOGGER.debug("IssueCertificate successful")
        return response


__all__ = ["CAKESServicer"]


def __server() -> None:
    from concurrent import futures

    cert, key = pskca.create_certificate_and_key(ca=True)
    ca = pskca.CA(cert, key, [cert])
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    pb2_grpc.add_CAKESServicer_to_server(  # type: ignore
        CAKESServicer(
            ca,
            unconditional_accept_ecdh,
            unconditional_accept_cert,
        ),
        server,
    )

    server.add_insecure_port("127.0.40.50:50052")
    print("starting server")
    server.start()
    server.wait_for_termination()
    print("server ended")


if __name__ == "__main__":
    __server()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cakes.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cakes.proto',
  package='CAKES',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\x0b\x63\x61kes.proto\x12\x05\x43\x41KES\"\x19\n\x07\x45\x43\x44HKey\x12\x0e\n\x06pubkey\x18\x01 \x01(\x0c\"\x05\n\x03\x41\x63k\"/\n\x17IssueCertificateRequest\x12\x14\n\x0c\x45ncryptedCSR\x18\x02 \x01(\x0c\"P\n\x15IssueCertificateReply\x12\x1b\n\x13\x45ncryptedClientCert\x18\x01 \x01(\x0c\x12\x1a\n\x12\x45ncryptedCertChain\x18\x02 \x01(\x0c\x32\xb7\x01\n\x05\x43\x41KES\x12,\n\x0c\x43lientPubkey\x12\x0e.CAKES.ECDHKey\x1a\n.CAKES.Ack\"\x00\x12,\n\x0cServerPubkey\x12\n.CAKES.Ack\x1a\x0e.CAKES.ECDHKey\"\x00\x12R\n\x10IssueCertificate\x12\x1e.CAKES.IssueCertificateRequest\x1a\x1c.CAKES.IssueCertificateReply\"\x00\x62\x06proto3'
)




_ECDHKEY = _descriptor.Descriptor(
  name='ECDHKey',
  full_name='CAKES.ECDHKey',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='pubkey', full_name='CAKES.ECDHKey.pubkey', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=47,
)


_ACK = _descriptor.Descriptor(
  name='Ack',
  full_name='CAKES.Ack',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=49,
  serialized_end=54,
)


_ISSUECERTIFICATEREQUEST = _descriptor.Descriptor(
  name='IssueCertificateRequest',
  full_name='CAKES.IssueCertificateRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='EncryptedCSR', full_name='CAKES.IssueCertificateRequest.EncryptedCSR', index=0,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=56,
  serialized_end=103,
)


_ISSUECERTIFICATEREPLY = _descriptor.Descriptor(
  name='IssueCertificateReply',
  full_name='CAKES.IssueCertificateReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='EncryptedClientCert', full_name='CAKES.IssueCertificateReply.EncryptedClientCert', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='EncryptedCertChain', full_name='CAKES.IssueCertificateReply.EncryptedCertChain', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=105,
  serialized_end=185,
)

DESCRIPTOR.message_types_by_name['ECDHKey'] = _ECDHKEY
DESCRIPTOR.message_types_by_name['Ack'] = _ACK
DESCRIPTOR.message_types_by_name['IssueCertificateRequest'] = _ISSUECERTIFICATEREQUEST
DESCRIPTOR.message_types_by_name['IssueCertificateReply'] = _ISSUECERTIFICATEREPLY
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ECDHKey = _reflection.GeneratedProtocolMessageType('ECDHKey', (_message.Message,), {
  'DESCRIPTOR' : _ECDHKEY,
  '__module__' : 'cakes_pb2'
  # @@protoc_insertion_point(class_scope:CAKES.ECDHKey)
  })
_sym_db.RegisterMessage(ECDHKey)

Ack = _reflection.GeneratedProtocolMessageType('Ack', (_message.Message,), {
  'DESCRIPTOR' : _ACK,
  '__module__' : 'cakes_pb2'
  # @@protoc_insertion_point(class_scope:CAKES.Ack)
  })
_sym_db.RegisterMessage(Ack)

IssueCertificateRequest = _reflection.GeneratedProtocolMessageType('IssueCertificateRequest', (_message.Message,), {
  'DESCRIPTOR' : _ISSUECERTIFICATEREQUEST,
  '__module__' : 'cakes_pb2'
  # @@protoc_insertion_point(class_scope:CAKES.IssueCertificateRequest)
  })
_sym_db.RegisterMessage(IssueCertificateRequest)

IssueCertificateReply = _reflection.GeneratedProtocolMessageType('IssueCertificateReply', (_message.Message,), {
  'DESCRIPTOR' : _ISSUECERTIFICATEREPLY,
  '__module__' : 'cakes_pb2'
  # @@protoc_insertion_point(class_scope:CAKES.IssueCertificateReply)
  })
_sym_db.RegisterMessage(IssueCertificateReply)



_CAKES = _descriptor.ServiceDescriptor(
  name='CAKES',
  full_name='CAKES.CAKES',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=188,
  serialized_end=371,
  methods=[
  _descriptor.MethodDescriptor(
    name='ClientPubkey',
    full_name='CAKES.CAKES.ClientPubkey',
    index=0,
    containing_service=None,
    input_type=_ECDHKEY,
    output_type=_ACK,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='ServerPubkey',
    full_name='CAKES.CAKES.ServerPubkey',
    index=1,
    containing_service=None,
    input_type=_ACK,
    output_type=_ECDHKEY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='IssueCertificate',
    full_name='CAKES.CAKES.IssueCertificate',
    index=2,
    containing_service=None,
    input_type=_ISSUECERTIFICATEREQUEST,
    output_type=_ISSUECERTIFICATEREPLY,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_CAKES)

DESCRIPTOR.services_by_name['CAKES'] = _CAKES

# @@protoc_insertion_point(module_scope)



from typing import List
import os
import random
import sys
import tempfile

import grpc
import pskca

from cryptography.x509 import Certificate
from cryptography.hazmat.primitives import serialization

import cakes
from cakes.proto import cakes_pb2_grpc
from concurrent import futures


import subprocess


def verify_cert(chain: List[Certificate], cert: Certificate) -> None:
    with tempfile.TemporaryDirectory() as x:
        signer_path = os.path.join(x, "signer.crt")
        cert_path = os.path.join(x, "cert.crt")
        with open(signer_path, "wb") as f:
            for c in chain:
                f.write(c.public_bytes(serialization.Encoding.PEM))
                f.write(b"\n")
        with open(cert_path, "wb") as f:
            f.write(cert.public_bytes(serialization.Encoding.PEM))
        subprocess.run(
            [
                "openssl",
                "verify",
                "-verbose",
                "-show_chain",
                "-trusted",
                signer_path,
                cert_path,
            ],
            capture_output=True,
            universal_newlines=True,
            check=True,
        )


def random_port() -> int:
    ports = list(range(1025, 65536))
    return random.choice(ports)


def random_ip() -> str:
    ips = list(range(0, 255))
    return "127.%s.%s.%s" % (
        random.choice(ips),
        random.choice(ips),
        random.choice(ips),
    )


def random_address() -> str:
    return random_ip() + ":" + str(random_port())


def test_protocol() -> None:
    address = random_address()
    print("Using address %s" % address, file=sys.stderr)
    cacert, key = pskca.create_certificate_and_key(ca=True)
    ca = pskca.CA(cacert, key, [cacert])
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    cakes_pb2_grpc.add_CAKESServicer_to_server(
        cakes.CAKESServicer(
            ca,
            cakes.unconditional_accept_ecdh,
            cakes.unconditional_accept_cert,
        ),
        server,
    )
    server.add_insecure_port(address)
    server.start()

    try:
        csr, unused_key = pskca.create_certificate_signing_request()
        with grpc.insecure_channel(address) as channel:
            client = cakes.CAKESClient(
                channel,
                csr,
                cakes.unconditional_accept_ecdh,
                cakes.unconditional_accept_cert,
            )
            clientcert, chain = client.run()
            try:
                verify_cert(chain, clientcert)
            except subprocess.CalledProcessError as e:
                assert 0, (e.stderr, e.stdout)

    finally:
        server.stop(0)


def test_separate_ca() -> None:
    address = random_address()
    print("Using address %s" % address, file=sys.stderr)
    cacert, cakey = pskca.create_certificate_and_key(ca=True)
    servercsr, serverkey = pskca.create_certificate_signing_request(
        cn="server",
    )
    servercert = pskca.issue_certificate(servercsr, cacert, cakey, ca=True)

    ca = pskca.CA(servercert, serverkey, [cacert, servercert])
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    cakes_pb2_grpc.add_CAKESServicer_to_server(
        cakes.CAKESServicer(
            ca,
            cakes.unconditional_accept_ecdh,
            cakes.unconditional_accept_cert,
        ),
        server,
    )
    server.add_insecure_port(address)
    server.start()

    try:
        clientcsr, unused_key = pskca.create_certificate_signing_request()
        with grpc.insecure_channel(address) as channel:
            client = cakes.CAKESClient(
                channel,
                clientcsr,
                cakes.unconditional_accept_ecdh,
                cakes.unconditional_accept_cert,
            )
            clientcert, chain = client.run()
            try:
                verify_cert(chain, clientcert)
            except subprocess.CalledProcessError as e:
                assert 0, (e.stderr, e.stdout)

    finally:
        server.stop(0)
